
/* Simple Video4Linux image grabber. Made for my Philips Vesta Pro
 * 
 * Copyright (C) 2000 geci, Larchmont, USA
 * eMail: devel@gecius.de
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at you option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public Licence for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA
 */

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <syslog.h>
#include <linux/types.h>
#include <linux/videodev.h>
#include <jpeglib.h>
#include <png.h>
#include <freetype.h>

#define QUAL_DEFAULT 75
#define VIDEO_DEV "/dev/video"
#define OUT_DEFAULT "/dev/stdout"
#define OUTFORMAT_DEFAULT 1		// 1=jpeg, 2=png
#define DEFAULT_FONT "/usr/X11R6/lib/X11/fonts/TrueType/Arialn.ttf"
#define DEFAULT_TIMESTAMP "%a, %e. %B %Y - %T"
#include "font.c"

int get_brightness_adj(unsigned char *image, long size, int *brightness) {
    long i, tot = 0;
    for (i=0;i<size*3;i++)
    tot += image[i];
    *brightness = (128 - tot/(size*3))/3;
    return !((tot/(size*3)) >= 126 && (tot/(size*3)) <= 130);
}

void usage (char *pname) {
    fprintf(stderr,
	"%s, Version %s\n"
	"Usage: %s [options]\n"
	" -h                This screen\n"
	" -l <seconds>      Loop forever with sleep for <seconds>\n"
	" -b                Disable vgrabbj's brightness adjustment\n"
	" -q <quality>      Quality setting (1-100, default: %d), JPEG only\n"
	" -o <jpeg|png>     Output format (default: jpeg)\n"
	" -f <filename>     Write to <filename> (default: stdout)\n"
	" -n                Disable the dot for each image (for daemon use)\n"
	" -d <device>       Read from <device> as input (default: %s)\n"
	" -s <device>       See capabilities of <device>\n"
	" -t <font-file>    Enable timestamp of image, needs full path to the font file\n"
	"                   (default: /usr/X11R6/lib/X11/fonts/TrueType/Arialn.ttf\n"
	" -p <format-str>   definable timestamp format (see man strftime)\n"
	"                   (default: \"%%a, %%e. %%B %%Y - %%T\")\n"
	"                   *MUST* be within \" and \" !\n"
	" -e                enable timestamp with defaults (default: disabled)\n"
	"                   if -t or -p is given, timestamp is already enabled\n"
	"\n"
	"Example: %s -l 5 -f /usr/local/image.jpg\n"
	"         Would write a single jpeg-image to image.jpg about every six seconds\n"
	"\n"
	"Currently, the video stream has to be rgb24. Sorry folks.\n",
	basename(pname), VERSION, basename(pname), QUAL_DEFAULT, VIDEO_DEV, basename(pname));
	exit (1);
}

void show_capabilities(char *in) {
    struct video_capability cap;
    struct video_window win;
    struct video_picture pic;
    int dev;
    struct jpeg_compress_struct cjpeg;
    struct jpeg_error_mgr jerr;
    
    dev = open(in, O_RDONLY);
    if (dev < 0) {
        fprintf(stderr, "Can't open device %s\n", in);
        exit(1);
    }
    if (ioctl(dev, VIDIOCGCAP, &cap) < 0) {
        perror("Get capabilities");
        close(dev);
        exit(1);
    }
    if (ioctl(dev, VIDIOCGPICT, &pic) < 0) {
        perror("Get picture properties");
        close(dev);
        exit(1);
    }
    if (ioctl(dev, VIDIOCGWIN, &win) < 0) {
        perror("Get overlay values");
        close(dev);
        exit(1);
    }
    fprintf(stderr,"Videodevice name: %s (%s)\n"
		"Capabilities\n"
		"Type     : %d\tValues can be looked up at videodev.h\n"
		"Channels : %d\n"
		"Audio    : %d\n"
		"MaxWidth : %d\n"
		"MaxHeight: %d\n"
		"MinWidth : %d\n"
		"MinHeigth: %d\n"
		"\nCurrent Settings:\n"
		"Brightness: %d\n"
		"Hue       : %d\n"
		"Color     : %d\n"
		"Contrast  : %d\n"
		"Whiteness : %d\n"
		"Depth     : %d\n"
		"Palette   : %d\tValues can be looked up at videodev.h\n"
		"Width     : %d\n"
		"Height    : %d\n"
		"Chromakey : %d\n",
		 in, cap.name, cap.type, cap.channels, cap.audios,
		 cap.maxwidth, cap.maxheight, cap.minwidth, cap.minheight,
		 pic.brightness, pic.hue, pic.colour, pic.contrast,
		 pic.whiteness, pic.depth, pic.palette, win.width, win.height,
		 win.chromakey);
    close(dev);
    exit(0);
}

int write_jpeg(char *buffer, FILE *x, int quality, int width, int height) {
    char *line;
    int n, y=0, i, line_width;

    struct jpeg_compress_struct cjpeg;
    struct jpeg_error_mgr jerr;
    JSAMPROW row_ptr[1];

    line=malloc(width * 3);
    if (!line) {
	perror("OUT OF MEMORY, Exiting...");
	syslog(LOG_ALERT, "OUT OF MEMORY, Exiting...");
	exit(1);
    }
    cjpeg.err = jpeg_std_error(&jerr);
    jpeg_create_compress (&cjpeg);
    cjpeg.image_width = width;
    cjpeg.image_height= height;
    cjpeg.input_components = 3;
    cjpeg.in_color_space = JCS_RGB;

    jpeg_set_defaults (&cjpeg);
    jpeg_set_quality (&cjpeg, quality, TRUE);
    cjpeg.dct_method = JDCT_FASTEST;

    jpeg_stdio_dest (&cjpeg, x);
    jpeg_start_compress (&cjpeg, TRUE);
    row_ptr[0]=line;
    line_width=width * 3;
    n=0;

    for (y = 0; y < height; y++) {
        for (i = 0; i< line_width; i+=3) {
	    line[i]   = buffer[n+2];
	    line[i+1] = buffer[n+1];
	    line[i+2] = buffer[n];
	    n+=3;
	}
	jpeg_write_scanlines (&cjpeg, row_ptr, 1);
    }
    jpeg_finish_compress (&cjpeg);
    jpeg_destroy_compress (&cjpeg);
    free (line);
    return(0);
}

int write_png(char *image, FILE *x, int width, int height) {

    register int y;
    register char *p;
    png_infop info_ptr;
    png_structp png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING,
	NULL, NULL, NULL);
    if (!png_ptr)
        return(1);
    info_ptr = png_create_info_struct (png_ptr);
    if (!info_ptr)
        return(1);

    png_init_io (png_ptr, x);
    png_set_IHDR (png_ptr, info_ptr, width, height,
    	8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
	PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
    png_set_bgr (png_ptr);
    png_write_info (png_ptr, info_ptr);
    p = image;
    width *= 3;
    for (y = 0; y < height; y++) {
        png_write_row (png_ptr, p);
        p += width;
    }
    png_write_end (png_ptr, info_ptr);
    png_destroy_write_struct (&png_ptr, &info_ptr);
    return(0);
}


int main(int argc, char *argv[]) {
    int y, dev, f, n, bpp = 24;
    boolean loop=FALSE;
    boolean logstd=TRUE;
    boolean in_loop=FALSE;
    int quality= QUAL_DEFAULT;
    int err_count2;
    int sleeptime;
    char *in = VIDEO_DEV;
    int outformat = OUTFORMAT_DEFAULT;
    char *out = OUT_DEFAULT;
    char *line;
    unsigned char *buffer;
    unsigned int i, src_depth;

    FILE *x;

    struct video_capability cap;
    struct video_window win;
    struct video_picture vpic;
    boolean enable_timestamp = FALSE;
    boolean brightness=TRUE;
    char *font = DEFAULT_FONT;
    char *timestamp = DEFAULT_TIMESTAMP;
    int font_size = 12;
#define TS_MAX 128
    char ts_buff[TS_MAX+1];
    int ts_len;
    int border = 2;
    int blend = 60;
    int align = 1;
    time_t t;
    struct tm *tm;
    TT_Engine engine;
    TT_Face face;
    TT_Face_Properties properties;
    TT_Instance instance;
    TT_Glyph *glyphs = NULL;
    TT_Raster_Map bit;
    TT_Raster_Map sbit;

    while ((n = getopt (argc, argv, "l:f:q:hd:ns:o:t:p:eb"))!=EOF) {
	switch (n) {
	    case 'l':
		loop = TRUE;
		sscanf (optarg, "%d", &sleeptime);
		break;
	    case 'f':
		out = optarg;
		break;
	    case 'q':
		sscanf (optarg, "%d", &quality);
		break;
	    case 'o':
		if (strstr(optarg,"j"))
		    outformat=1;
		if (strstr(optarg,"n"))
		    outformat=2;
		break;
	    case 'd':
		in = optarg;
		break;
	    case 'n':
		logstd=FALSE;
		break;
	    case 't':
		font=optarg;
		enable_timestamp=TRUE;
		break;
	    case 'p':
		timestamp=optarg;
		enable_timestamp=TRUE;
		break;
	    case 'e':
		enable_timestamp=TRUE;
		break;
	    case 'b':
		brightness=FALSE;
		break;
	    case 's':
		show_capabilities(optarg);
	    default:
		usage (argv[0]);
		break;
	}
    }

    if (loop) 
	syslog(LOG_INFO, "Reading images from %s\n", in);
    else
	fprintf(stderr, "Reading image from %s\n", in);

    do {
	if (!in_loop) {
    	    dev = open(in, O_RDONLY);
	}
        if (dev < 0) {
	    if (loop) {
		syslog(LOG_ERR, "Problem opening %s\n", in);
	        usleep (250000);
		in_loop=FALSE;
	    }
	    else {
	        fprintf(stderr, "Can't open device %s\n", in);
	        exit(1);
	    }
	}
	else {
	    in_loop=TRUE;
	    if (ioctl(dev, VIDIOCGCAP, &cap) < 0) {
		if (loop) {
		    syslog(LOG_WARNING, "Problem getting video capabilities %m\n");
		    usleep(250000);
		}
		else {
		    perror("Problem getting video capabilities");
    		    close(dev);
	    	    exit(1);
		}
	    }
	    else {
	        if (ioctl(dev, VIDIOCGWIN, &win) <0) {
	            if (loop) {
			syslog(LOG_WARNING, "Problem setting window overlay %m\n");
		        usleep (250000);
		    }
		    else {
        	        perror("Problem setting window overlay");
        	        close(dev);
        	        exit(1);
		    }
		}
		else {
		    if (ioctl(dev, VIDIOCGPICT, &vpic) < 0) {
			if (loop) {
			    syslog(LOG_WARNING, "Problem getting picture properties %m\n");
			    usleep(250000);
			}
			else {
        		    perror("Problem getting picture properties");
        		    close(dev);
        		    exit(1);
			}
		    }
		    else {
			vpic.depth=24;
    			vpic.palette=VIDEO_PALETTE_RGB24;

			buffer = malloc(win.width * win.height * bpp);
			if (!buffer) {
			    perror("Out of Memory!");
			    syslog(LOG_ALERT, "Out of Memory! Exiting...\n");
			    close(dev);
			    exit(9);
			}
			err_count2=0;
			if (brightness) {
			    do {
				int newbright;
				read(dev, buffer, win.width * win.height * bpp);
				f = get_brightness_adj(buffer, win.width * win.height, &newbright);
				if (f) {
				    vpic.brightness += (newbright << 8);
				    if(ioctl(dev, VIDIOCSPICT, &vpic)==-1) {
					perror("Problem setting brightness");
					syslog(LOG_DEBUG, "Problem setting brightness %m\n");
					break;
				    }
				    err_count2++;
				    if (err_count2>100) {
					syslog(LOG_WARNING,"Brightness not optimal\n");
					perror("Brightness not optimal");
					break;
				    }
				}
			    } while (f);
			}
			else {
			read(dev, buffer, win.width * win.height * bpp);
			}
			close(dev);
			in_loop=FALSE;
			if (font && timestamp && enable_timestamp) {
			    if (TT_Init_FreeType (&engine)) {
				font = NULL;
			    }
			    if (Face_Open (font, engine, &face, &properties, &instance, font_size)){
				if (loop) {
				    syslog(LOG_WARNING, "Font not found: %s\n", font);
				}
				else {
				    fprintf(stderr,"Font not found: %s\n",font);
				}
				TT_Done_FreeType (engine);
				font = NULL;
			    }
			}

			if (font && timestamp && enable_timestamp) {
			    time (&t);
			    tm = localtime (&t);
			    ts_buff[TS_MAX] = '\0';
			    strftime (ts_buff, TS_MAX, timestamp, tm);
			    ts_len = strlen (ts_buff);

			    glyphs = Glyphs_Load (face, &properties, instance, ts_buff, ts_len);
			    Raster_Init(face, &properties,instance,ts_buff,ts_len, border, glyphs, &bit);
			    Raster_Small_Init (&sbit, &instance);
			    Render_String (glyphs, ts_buff, ts_len, &bit, &sbit, border);
			    if (bit.bitmap) {
				int x, y, psize, i, x_off, y_off;
				unsigned char *p;

				psize = 3;

				switch (align) {
				    case 1:
					x_off = (win.width - bit.width) * psize;
					y_off = 0;
					break;
				    case 2:
					x_off = 0;
					y_off = win.height - bit.rows;
					break;
				    case 3:
					x_off = (win.width - bit.width) * psize;
					y_off = win.height - bit.rows;
					break;
				    default:
					x_off = y_off = 0;
					break;
				}

				for (y = 0; y < bit.rows; y++) {
				    p = buffer + (y + y_off) * (win.width * psize) + x_off;
				    for (x = 0; x < bit.width; x++) {
					switch (((unsigned char *)bit.bitmap)
					    [((bit.rows-y-1)*bit.cols)+x]) {
					    case 0:
						for (i = 0; i < psize; i++) {
						    *p = (255 * blend + *p * (100 - blend))/100;
						    p++;
						}
						break;
					    case 1:
					    	for (i = 0; i < psize; i++) {
						    *p = (220 * blend + *p * (100 - blend))/100;
						    p++;
						}
						break;
					    case 2:
						for (i = 0; i < psize; i++) {
						    *p = (162 * blend + *p * (100 - blend))/100;
						    p++;
						}
						break;
					    case 3:
						for (i = 0; i < psize; i++) {
						    *p = (64 * blend + *p * (100 - blend))/100;
						    p++;
						}
						break;
					    default:
						for (i = 0; i < psize; i++) {
						    *p = (0 * blend + *p * (100 - blend))/100;
						    p++;
						}
						break;
					}
				    }
				}
			    }
			    Raster_Done (&sbit);
			    Raster_Done (&bit);
			    Glyphs_Done (glyphs);
			    glyphs = NULL;
			}

			if (font && timestamp && enable_timestamp) {
			    Face_Done (instance, face);
			    TT_Done_FreeType (engine);
			}

			do {
			    x = fopen(out, "w+");
			    if (!x) {
				if (loop) {
				    syslog(LOG_WARNING, "Could not open outputfile: %s\n", out);
				    usleep(250000);
				}
				else {
				    perror("Could not open Outputfile:");
				    close(dev);
				    exit(1);
				}
			    }
			} while (!x);
			switch (outformat) {
			case 1:
			    while (write_jpeg(buffer, x, quality, win.width, win.height)) {
				syslog(LOG_WARNING, "Could not write to outputfile: %s\n", out);
				usleep(25000);
			    }
			    break;
			case 2:
			    write_png(buffer, x, win.width, win.height);
			    break;
			default:
			    usage(argv[0]);
			    break;
			}
			free (buffer);
			if (loop) {
			    if (logstd) {
				fprintf(stderr,".");
			    }
			    sleep(sleeptime);
			}
		    }
		}
	    }
	}
    } while (loop);
}
